from pathlib import Path

import pytest

from hyprshade.cli.ls import ls_dirs
from hyprshade.cli.utils import write_systemd_user_unit


class TestLsDirs:
    def test_empty(self):
        assert list(ls_dirs([])) == []

    def test_single(self, tmp_path: Path):
        (tmp_file := tmp_path / "foo").touch()
        assert list(ls_dirs([tmp_path])) == [str(tmp_file)]

    def test_multiple(self, tmp_path: Path):
        (tmp_file1 := tmp_path / "bar").touch()
        (tmp_file2 := tmp_path / "foo").touch()
        assert list(ls_dirs([tmp_path])) == list(map(str, [tmp_file1, tmp_file2]))

    def test_multiple_dirs(self, tmp_path_factory: pytest.TempPathFactory):
        paths = []
        for dir_ in ["foo", "bar", "baz"]:
            tmp_path_ = tmp_path_factory.mktemp(dir_)
            (tmp_path_ / f"qux{dir_}").touch()
            paths.append(tmp_path_)

        assert list(ls_dirs(paths)) == list(
            map(str, [paths[1] / "quxbar", paths[2] / "quxbaz", paths[0] / "quxfoo"])
        )


def test_write_systemd_user_unit(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    (tmp_path / "systemd" / "user").mkdir(parents=True)

    write_systemd_user_unit("service", "foo")
    assert (tmp_path / "systemd" / "user" / "hyprshade.service").read_text() == "foo"

    write_systemd_user_unit("timer", "bar")
    assert (tmp_path / "systemd" / "user" / "hyprshade.timer").read_text() == "bar"



from collections.abc import Callable

import pytest
from click.shell_completion import CompletionItem

from hyprshade.cli.utils import ShaderParamType
from tests.types import ShaderPathFactory

pytestmark = [pytest.mark.usefixtures("_clear_shader_env")]


@pytest.fixture
def _shaders(shader_path_factory: ShaderPathFactory):
    shader_path_factory("foo")
    shader_path_factory("bar")


ShaderComplete = Callable[[str], list[CompletionItem]]


@pytest.fixture
def shader_complete() -> ShaderComplete:
    shader_param = ShaderParamType()

    def _shader_complete(incomplete: str):
        return shader_param.shell_complete(None, None, incomplete)  # type: ignore[arg-type]

    return _shader_complete


def test_shell_complete_empty(shader_complete: ShaderComplete):
    assert shader_complete("foo") == []


@pytest.mark.usefixtures("_shaders")
def test_shell_complete(shader_complete: ShaderComplete):
    completion_items = shader_complete("foo")

    for expected_value, item in zip(["bar", "foo"], completion_items, strict=True):
        assert item.value == expected_value


def test_shell_complete_path(shader_complete: ShaderComplete):
    completion_items = shader_complete("./")

    for expected_value, item in zip(["./"], completion_items, strict=True):
        assert item.value == expected_value



import re
from pathlib import Path

import pytest
from click.testing import CliRunner

from hyprshade.cli import cli
from hyprshade.shader import hyprctl
from tests.types import ShaderPathFactory

pytestmark = [
    pytest.mark.requires_hyprland(),
    pytest.mark.usefixtures("_save_screen_shader"),
    pytest.mark.usefixtures("_clear_shader_env"),
    pytest.mark.usefixtures("_clear_screen_shader"),
]


@pytest.mark.parametrize("is_long", [False, True])
def test_empty(is_long: bool, runner: CliRunner):
    cmd_args = ["--long"] if is_long else []
    result = runner.invoke(cli, ["ls", *cmd_args])

    assert result.exit_code == 0
    assert result.output == ""


@pytest.mark.parametrize("is_long", [False, True])
def test_single(is_long: bool, runner: CliRunner, shader_path_env: Path):
    cmd_args = ["--long"] if is_long else []
    result = runner.invoke(cli, ["ls", *cmd_args])

    assert result.exit_code == 0

    pattern = "shader" + (
        rf" +{re.escape(shader_path_env.parent.as_posix())}" if is_long else ""
    )
    assert re.match(pattern, result.output.strip()) is not None


@pytest.mark.parametrize("is_long", [False, True])
def test_multiple(
    is_long: bool, runner: CliRunner, shader_path_factory: ShaderPathFactory
):
    shader_names = ["shader1", "shader2", "shader3"]
    shader_paths = list(map(shader_path_factory, shader_names))

    cmd_args = ["--long"] if is_long else []
    result = runner.invoke(cli, ["ls", *cmd_args])

    assert result.exit_code == 0

    for name, path, line in zip(
        shader_names, shader_paths, result.output.strip().splitlines(), strict=True
    ):
        pattern = re.escape(name) + (
            rf" +{re.escape(path.parent.as_posix())}" if is_long else ""
        )
        assert re.match(pattern, line.strip()) is not None


@pytest.mark.parametrize("is_long", [False, True])
@pytest.mark.parametrize("current_index", [0, 1, 2])
def test_active(
    is_long: bool,
    current_index: int,
    runner: CliRunner,
    shader_path_factory: ShaderPathFactory,
):
    shader_names = ["shader1", "shader2", "shader3"]
    shader_paths = list(map(shader_path_factory, shader_names))

    hyprctl.set_screen_shader(shader_paths[current_index].as_posix())
    cmd_args = ["--long"] if is_long else []
    result = runner.invoke(cli, ["ls", *cmd_args])

    assert result.exit_code == 0

    for i, (name, path, line) in enumerate(
        zip(shader_names, shader_paths, result.output.strip().splitlines(), strict=True)
    ):
        pattern = (
            (re.escape("* ") if i == current_index else "")
            + re.escape(name)
            + (rf" +{re.escape(path.parent.as_posix())}" if is_long else "")
        )
        assert re.match(pattern, line.strip()) is not None



from __future__ import annotations

import os
import tomllib
from os import path
from typing import cast

from more_itertools import first_true, nth, partition

from .types import ConfigDict, DefaultShadeConfig, ShaderConfig
from .utils import hypr_config_home, hyprshade_config_home


class Config:
    _config_path: str
    _dict: ConfigDict

    def __init__(self, path_: str | None = None):
        path_ = path_ or Config.get_path()
        if path_ is None:
            raise FileNotFoundError(
                "Could not find a config file; see https://github.com/loqusion/hyprshade#scheduling"
            )
        self._config_path = path_
        self._dict = Config._load(path_)
        self._validate()

    def partition(self) -> tuple[list[ShaderConfig], DefaultShadeConfig | None]:
        no_default, yes_default = partition(
            lambda s: s.get("default", False), self._dict["shades"]
        )
        rest = cast(list[ShaderConfig], list(no_default))
        default = cast(DefaultShadeConfig, nth(yes_default, 0))

        assert nth(yes_default, 0) is None, "There should be only one default shade"

        return rest, default

    def _validate(self) -> None:
        shaders = self._dict.get("shades", [])
        if not isinstance(shaders, list):
            raise ConfigError(self._config_path, "`shades` must be a list")
        for shader in shaders:
            if not shader.get("name"):
                raise ConfigError(
                    self._config_path, "`name` is required for each item in `shades`"
                )
            if not shader.get("start_time") and shader.get("default") is not True:
                raise ConfigError(
                    self._config_path,
                    f"Non-default shader '{shader['name']}' must define `start_time`",
                )

    @staticmethod
    def _load(path_: str) -> ConfigDict:
        with open(path_, "rb") as f:
            return cast(ConfigDict, tomllib.load(f))

    @staticmethod
    def get_path() -> str | None:
        candidates = [
            os.getenv("HYPRSHADE_CONFIG"),
            path.join(hypr_config_home(), "hyprshade.toml"),
            path.join(hyprshade_config_home(), "config.toml"),
        ]
        return first_true((c for c in candidates if c is not None), pred=path.isfile)


class ConfigError(Exception):
    def __init__(self, config_path: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config_path = config_path

    def __str__(self) -> str:
        msg = super().__str__()
        return f"Failed to parse {self.config_path}: {msg}"
